﻿//*****************************************************************************
// Description     : Main sample program to measure histogram
// Target Device   : APN504X
// Target Platform : Windows, .NET Framework, C# (Console Application)
// Date            : August 2017
// Version         : 1.0
//*****************************************************************************
using System;
using System.Threading;

namespace APN504X_sample
{
	using static Device;
	using static Files;
	using DEVICE_COM_INFO = SiTCP.SiTCPcom;

	class Program
	{
		static int Main(string[] args)
		{
			int run_ready;
			int ret;
			DEVICE_COM_INFO comInfo;
			ulong meas_count;

			run_ready = 0;

			// 1. connect to the device
			ret = device_connect(out comInfo);
			if (ret != 0) {
				return(1);
			}

			// 2. set parameters to the device
			meas_count = device_msec2count(Defines.MEASUREMENT_TIME);	// convert msec to count
			ret = config_device(comInfo, meas_count);
			if (ret == 0) {
				run_ready = 1;
			} else {
				Console.WriteLine("setting device config error({0})...", ret);
			}

			if (run_ready == 1) {	// ready to run
				// 3. start measurement
				Console.WriteLine("start measurement!!!");
				ret = run_measurement(comInfo, meas_count);

				Console.WriteLine("finish measurement...");

				if (ret >= 0) {
					Console.WriteLine("measurement completed!!!");
				}
				else {
					Console.WriteLine("measurement error({0})...\n", ret);
				}
			}

			// 4. post processing
			device_disconnect(comInfo);

			return(0);
		}


		//#########################################################################
		//  Name      : config_device
		//  Function  : Set parameters to device
		//  Arguments :
		//      in comInf   : instance of DEVICE_COM_INFO class
		//      in ms_count : measurement time (counts)
		//  Return :
		//       0 : success
		//      -n : failed at nth parameter
		//  Note : 
		//#########################################################################
		static int config_device(DEVICE_COM_INFO comInf, ulong ms_count)
		{
			int i, ret;
			Command cmd_ref = new Command();

			// set for each CH
			for (i = 0; i < Device.MAX_CHNUM; i++) {
                int chn = i + 1;

				// polarity: positive
				ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_POL", chn), (ushort)cmd_ref.GetVal("PARAM_POL_POS"));
				if (ret < 0)  { return(-1); }

				// ADC gain: 4096
				ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_ADG", chn), (ushort)cmd_ref.GetVal("PARAM_ADG_4096"));
				if (ret < 0)  { return(-2); }

				// analog coarse gain: x5
				ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_ACG", chn), (ushort)cmd_ref.GetVal("PARAM_ACG_x5"));
				if (ret < 0)  { return(-3); }

				// fast trigger threshold: 50
				ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_FTH", chn), 50);
				if (ret < 0)  { return(-4); }

				// fast filter pole zero: 0
				ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_FPZ", chn), 0);
				if (ret < 0)  { return(-5); }

				// slow filter rise time: 800ns(80 @1digit=10ns)
				ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_SFR", chn), 80);
				if (ret < 0)  { return(-6); }

				// slow filter peaking time: 1000ns(100 @1digit=10ns)
				ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_SFP", chn), 100);
				if (ret < 0)  { return(-7); }

				// slow filter threshold: 40
				ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_STH", chn), 40);
				if (ret < 0)  { return(-8); }

				// slow filter pole zero: 65ns
				ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_SPZ", chn), 65);
				if (ret < 0)  { return(-9); }

				// digital coarse gain: x8
				ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_DCG", chn), (ushort)cmd_ref.GetVal("PARAM_DCG_x8"));
				if (ret < 0)  { return(-10); }

				// digital fine gain: x0.5 (8193 x 0.5 - 2 = 4094)
				ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_DFG", chn), 4094);
				if (ret < 0)  { return(-11); }

                // analog fine gain: 200
                ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_FGD", chn), 200);
                if (ret < 0) { return (-11); }

                // analog polezero: 0
                ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_PZD", chn), 0);
                if (ret < 0) { return (-101); }

                // reset filter: 0,1,0
                ret = device_write_3step(comInf, cmd_ref.GetVal("COMMAND_RFL", chn));
				if (ret < 0)  { return(-102); }
			}

			// DAC monitor: CH1 slow
			ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_DAC"), (ushort)cmd_ref.GetVal("PARAM_DAC_CH1_SLOW"));
			if (ret < 0)  { return(-13); }

			// measurement mode: histogram
			ret = device_write_16bit(comInf, cmd_ref.GetVal("COMMAND_MOD"), (ushort)cmd_ref.GetVal("PARAM_MOD_HIST"));
			if (ret < 0)  { return(-14); }

			// measurement time
			ret = device_write_48bit(comInf, cmd_ref.GetVal("COMMAND_MTM"), ms_count);
			if (ret < 0)  { return(-15); }

			return(0);
		}


		//#########################################################################
		//  Name      : run_measurement
		//  Function  : Measurement processing
		//  Arguments :
		//      in comInf   : instance of DEVICE_COM_INFO class
		//      in ms_count : measurement time (counts)
		//  Return :
		//       0 : success
		//      -1 : failed while measurement
		//      -2 : failed to write file header
		//      -3 : failed to receive histogram data
		//  Note : 
		//#########################################################################
		static int run_measurement(DEVICE_COM_INFO comInf, ulong ms_count)
		{
			int ret;
			ulong now_count = 0;
			uint now_msec;
			ushort status = 0;
			uint[] input_cps = new uint[Device.MAX_CHNUM];
			uint[] throughput_cps = new uint[Device.MAX_CHNUM];
			Command cmd_ref = new Command();

			// clear data
			device_write_3step(comInf, cmd_ref.GetVal("COMMAND_CLR"));
			// wait for completion of clear processing
			Thread.Sleep(100);

			// start measurement
			device_start_measurement(comInf);

			// display top 2 lines of status field to explain each column
			Console.WriteLine("<time(msec)>   |        <input rate(cps) / throughput rate(cps)>");
			Console.WriteLine(" real/total    |     CH1     |     CH2     |     CH3     |     CH4");
			Console.WriteLine("---------------+-------------+-------------+-------------+-------------");

			while(true) {
				// read status
				ret = device_read_status(comInf,
										ref status, ref now_count,
										ref input_cps, ref throughput_cps);

				if (ret < 0) { // communication error
					// quit measurement
					break;
				}
				// status is 'stopped' or measurement time elapsed
				if ( (status == cmd_ref.GetVal("PARAM_AQS_STOP")) || (now_count >= ms_count) ) {
					break;
				}

				Thread.Sleep(200);	// wait for 200ms
			}

			// stop measurement
			device_stop_measurement(comInf);

			// quit by error
			if (ret < 0) {
				return(-1);
			}

			// write file header
			now_msec = (uint)(device_count2msec(now_count));	// convert count to msec
			ret = file_write_header(Defines.SAVE_FILENAME, now_msec, input_cps, throughput_cps);
			if (ret < 0) {
				return(-2);
			}

			// receive histogram data then write out to file
			ret = receive_data(comInf);
			if (ret < 0) {
				return(-3);
			} else {
				return(0);
			}
		}


		//#########################################################################
		//  Name      : receive_data
		//  Function  : receiving collected data
		//  Arguments :
		//      in comInf : instance of DEVICE_COM_INFO class
		//  Return :
		//       0 : success
		//      -1 : failed to receive data
		//      -2 : failed to write file
		//  Note : 
		//#########################################################################
		static int receive_data(DEVICE_COM_INFO comInf)
		{
			int i;
			int ret = 0;
			int result = 0;
			byte[][] hdata = new byte[Device.MAX_CHNUM][];

			// receive histogram data
			for (i = 0; i < Device.MAX_CHNUM; i++) {
				hdata[i] = new byte[Defines.NUM_DATA * 4];
				ret = device_read_histogram(comInf, ref hdata[i], Defines.NUM_DATA * 4,
											SiTCP.DEVICE_TIMEOUT_DATA, i);
			}

			if (ret > 0) {
				// write data to file
				ret = file_write_data(Defines.SAVE_FILENAME, Device.MAX_CHNUM, hdata);
				if (ret < 0) {
					// failed to write
					result = -2;
				}
			} else {
				// failed to receive data
				result = -1;
			}

			return(result);
		}

	} // end of class
} // end of namespace

