﻿//*****************************************************************************
// Description     : Class for controlling device
// Target Device   : APN504X
// Target Platform : Windows, .NET Framework, C#
// Date            : Auguts 2017
// Version         : 1.0
//*****************************************************************************
using System;
using System.Net;

namespace APN504X_sample
{
	using static SiTCP;
	using DEVICE_COM_INFO = SiTCP.SiTCPcom;

	public static class Device
	{
		//****************************** constant value definition ********************
		private const ulong DEVICE_CNT_PER_NSEC = (10);   // 1count = 10nsec

		public const int MAX_CHNUM = 4;				// number of CH
		public const uint OFFSETADDR_CH = 0x0200;	// command address offset per CH

		//****************************** method aliases ******************************
		public static int device_connect(out DEVICE_COM_INFO comRef)
		{
			return sitcp_connect(out comRef);
		}
		public static void device_disconnect(DEVICE_COM_INFO comRef)
		{
			sitcp_disconnect(comRef);
		}
		public static int device_write_16bit(DEVICE_COM_INFO comRef, uint cmd_addr, ushort wdata)
		{
			return sitcp_write_16bit(comRef, cmd_addr, wdata);
		}
		public static int device_write_48bit(DEVICE_COM_INFO comRef, uint cmd_addr, ulong wdata)
		{
			return sitcp_write_48bit(comRef, cmd_addr, wdata);
		}

		//****************************** method definition ******************************
		// stop measurement
		public static void device_stop_measurement(DEVICE_COM_INFO pCmd)
		{
			Command cmd_ref = new Command();
			sitcp_write_16bit( pCmd, cmd_ref.GetVal("COMMAND_AQS"), 
									(ushort)cmd_ref.GetVal("PARAM_AQS_STOP") );
		}

		// convert msec to count
		public static ulong device_msec2count(ulong msec)
		{
			return ( ((ulong)msec * 1000000UL) / DEVICE_CNT_PER_NSEC );
		}
		// convert count to msec
		public static ulong device_count2msec(ulong count)
		{
			return ( ((ulong)count * DEVICE_CNT_PER_NSEC) / 1000000UL );
		}

		//#########################################################################
		//  Name      : device_write_3step
		//  Function  : issue command 3 times with value of 0, 1, 0 respectively
		//  Arguments :
		//      in comInf   : instance of DEVICE_COM_INFO
		//      in cmd_addr : command address
		//  Return :
		//       0 : success
		//      -n : failed at nth command
		//  Note : 
		//#########################################################################
		public static int device_write_3step(DEVICE_COM_INFO comInf, uint cmd_addr)
		{
			int ret;

			// issue command(1st time value is 0)
			ret = sitcp_write_16bit(comInf, cmd_addr, 0);
			if (ret < 0) {
				return(-1);
			}

			// issue command(2nd time value is 1)
			ret = sitcp_write_16bit(comInf, cmd_addr, 1);
			if (ret < 0) {
				return(-2);
			}

			// issue command(3rd time value is 0)
			ret = sitcp_write_16bit(comInf, cmd_addr, 0);
			if (ret < 0) {
				return(-3);
			}

			return(0);
		}


		//#########################################################################
		//  Name      : device_start_measurement
		//  Function  : start measurement
		//  Arguments :
		//      in comInf : instance of DEVICE_COM_INFO
		//  Return :
		//       0 : success
		//      -1 : failed at stop command
		//      -2 : failed at start command
		//  Note : 
		//#########################################################################
		public static int device_start_measurement(DEVICE_COM_INFO comInf)
		{
			int ret;
			Command cmd_ref = new Command();

			// issue stop command first, in case of in the middle of measurement
			ret = sitcp_write_16bit(comInf, cmd_ref.GetVal("COMMAND_AQS"),
											(ushort)cmd_ref.GetVal("PARAM_AQS_STOP") );
			if (ret < 0) {
				return(-1);
			}

			// issue start command
			ret = sitcp_write_16bit(comInf, cmd_ref.GetVal("COMMAND_AQS"),
											(ushort)cmd_ref.GetVal("PARAM_AQS_START") );
			if (ret < 0) {
				return(-2);
			}

			return(0);
		}


		//#########################################################################
		//  Name      : device_read_status
		//  Function  : read some status information from device
		//  Arguments :
		//      in  comInf    : instance of DEVICE_COM_INFO
		//      out pStat     : pointer to the measurement status
		//      out pTime     : pointer to the elapsed time
		//      out pInputCps : pointer to the input count rate
		//      out pThrouCps : pointer to the throughput count rate
		//  Return :
		//       0 : success
		//      -n : failed at nh command
		//  Note : 
		//#########################################################################
		public static int device_read_status(DEVICE_COM_INFO comInf,
											ref ushort pStat, ref ulong pTime,
				 							ref uint[] pInputCps, ref uint[] pThrouCps)
		{
			int ret, i;
			uint now_time;  // elapsed time(msec)
			Command cmd_ref = new Command();

			//
			// read each status
			// (read all data before display
			// in order to obtain values at the same possible point in time)
			//

			// measurement status: 2byte value(receive 16bit length data)
			pStat = sitcp_read_16bit(comInf, cmd_ref.GetVal("COMMAND_AQS"), out ret);
			if (ret != 0) { return (-1); }

			// input count rate: 4byte value(receive 32bit length data)
			for (i = 0; i < MAX_CHNUM; i++) {
				pInputCps[i] = sitcp_read_32bit(comInf, (uint)(cmd_ref.GetVal("STATUS_ICR") + OFFSETADDR_CH * i), out ret);
			}
			if (ret != 0) { return (-2); }

			// throughput count rate: 4byte value(receive 32bit length data)
			for (i = 0; i < MAX_CHNUM; i++) {
				pThrouCps[i] = sitcp_read_32bit(comInf, (uint)(cmd_ref.GetVal("STATUS_TCR") + OFFSETADDR_CH * i), out ret); 
			}
			if (ret != 0) { return (-3); }

			// elapsed time: 6byte value(receive 48bit length data)
			pTime = sitcp_read_48bit(comInf, cmd_ref.GetVal("STATUS_RTM"), out ret);
			if (ret != 0) { return (-4); }

			//
			// display(only for Debug output window)
			//
			now_time = (uint)device_count2msec(pTime);    // convert count to msec
			Console.Write("{0,5:D}/{1,5:D}   ", now_time, (uint)Defines.MEASUREMENT_TIME);
			for (i = 0; i < MAX_CHNUM; i++) {
				Console.Write(" | {0,5:D}/{1,5:D}", pInputCps[i], pThrouCps[i]);
			}
			Console.WriteLine();

			return(0);
		}


		//#########################################################################
		//  Name      : device_read_histogram
		//  Function  : read histogram data from device
		//  Arguments :
		//      in  comInf  : instance of DEVICE_COM_INFO
		//      out rbuffer : pointer to the data buffer
		//      in  nbyte   : request data size(byte)
		//      in  timeout : receive timeout value(msec)
		//      in  chidx   : CH index(0 started)
		//  Return :
		//       0 : received data size
		//      -1 : failed to data request
		//      -2 : failed to data receive
		//  Note : this function converts histogram data from network byte order
		//         to host byte order
		//#########################################################################
		public static int device_read_histogram(DEVICE_COM_INFO comInf,
												ref byte[] rbuffer, int nbyte, int timeout, int chidx)
		{
			int i, unit;
			int ret;
			int result;
			int wtimeout;
			int read_byte, recv_byte;
			Command cmd_ref = new Command();
			int preConv, postConv;

			result = 0;

			// request histogram data of CH(specified by chidx)
			ret = sitcp_write_16bit(comInf, cmd_ref.GetVal("COMMAND_RQH"), (ushort)chidx);

			if (ret >= 0) {  // succeeded to request histogram data
				recv_byte = 0;
				wtimeout = timeout;

				while(true) {
					// update request data size
					read_byte = nbyte - recv_byte;

					// receive histogram data via port for fast data transfer
					ret = sitcp_recv_data(comInf, ref rbuffer, recv_byte, read_byte, wtimeout);

					if (ret <= 0) {
						// no data or receive error
						break;
					}

					// update received data size
					recv_byte += ret;

					if (recv_byte >= nbyte) {
						// finish when received data size reached to request one
						break;
					}

					wtimeout = 10;   // timeout value is 10ms after the second receiving
				}

				if (recv_byte > 0) { // succeeded to receive
					// set total received size
					result = recv_byte;

					// convert histogram data from network byte order to host byte order
					unit = nbyte / sizeof(uint);
					for(i = 0; i < unit; i++) {
						preConv = BitConverter.ToInt32(rbuffer, i*4);
						postConv = IPAddress.NetworkToHostOrder(preConv);
						Array.Copy( BitConverter.GetBytes(postConv), 0, rbuffer, i*4, 4);
					}
				}
				else { // failed to receive
					result = -2;
				}
			}
			else { // failed to request
				result = -1;
			}

			return(result);
		}

	} // end of class
} // end of namespace
