﻿//*****************************************************************************
// Description     : Class for SiTCP communication
// Target Device   : APN504X
// Target Platform : Windows, .NET Framework, C#
// Date	           : August 2017
// Version         : 1.0
//*****************************************************************************
using System;
using System.Threading;
using System.Runtime.InteropServices;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Linq;

namespace APN504X_sample
{
	public static class SiTCP
	{
		//******************** private constant value definition ********************
		// Size of the buffer of socket
		private const int DEVICE_SOCKBUF_CMD = (16 * 1024);		// for command communication(16kB)
		private const int DEVICE_SOCKBUF_DATA = (1024 * 1024);	// for receiving data(1MB)
		// Timeout value
		private const int DEVICE_TIMEOUT_CMD = 200;		// for waiting command response
		public const int DEVICE_TIMEOUT_DATA = 200;	// for waiting data

		// command communication: header common part
		private const byte SITCP_VER_TYPE = 0xff;		// version, type

		// command communication: read command flag
		private const byte SITCP_READ_REQ_CMDFLAG = 0xc0;	// request
		private const byte SITCP_READ_ACK_CMDFLAG = 0xc8;	// acknowledge

		// command communication: write command flag
		private const byte SITCP_WRITE_REQ_CMDFLAG = 0x80;	// request
		private const byte SITCP_WRITE_ACK_CMDFLAG = 0x88;	// acknowledge

		// maximum size of data in SITCP_FORMAT
		private const int MAX_CMD_BYTE = 256;

		//******************** structure type definition ********************
		// header for command communication 
		struct SITCP_HEADER
		{ 
			public byte ver_type;	// version(4bit), type(4bit)
			public byte cmd_flag;	// command(4bit), flag(4bit)
			public byte id;			// identifier(0-255)
			public byte length;		// data length(byte)
			public uint address;	// command address(32bit)
		}

		// data format for command communication 
		[StructLayout(LayoutKind.Explicit)]
		struct SITCP_FORMAT
		{
			[FieldOffset(0)]
			public SITCP_HEADER header;
 
			// belows are union(used one of them)
			[FieldOffset(8)]
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = MAX_CMD_BYTE)]
			public byte[] u08;	// unit of 8bit

			[FieldOffset(8)]
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = MAX_CMD_BYTE/2)]
			public ushort[] u16;	// unit of 16bit

			[FieldOffset(8)]
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = MAX_CMD_BYTE/4)]
			public uint[] u32;	// unit of 32bit

			[FieldOffset(8)]
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = MAX_CMD_BYTE/8)]
			public ulong[] u64;	// unit of 64bit
			// end of union
		}

		// Socket communication 
		public struct SiTCPcom
		{
			public UdpClient cmdClient;		// for command communication
			public TcpClient dataClient;	// for fast data transfer
		}


		//****************************** method definition ******************************
		//#########################################################################
		//  Name	  : sitcp_connect
		//  Function  : create UdpClient and TcpClient instance and connect to device
		//  Arguments :
		//	  out comRef : SiTCPcom class instance
		//  Return :
		//	   0 : success
		//	  -1 : failed to connect
		//  Note 
		//#########################################################################
		public static int sitcp_connect(out SiTCPcom comRef)
		{
			// create new instance of UdpClient and TcpClient
			UdpClient udpClient = new UdpClient(Defines.DEVICE_PORT_CTRL_OWNPC);
			TcpClient tcpClient = new TcpClient();

			// change the size of the receive buffer
			udpClient.Client.ReceiveBufferSize = DEVICE_SOCKBUF_CMD;
			tcpClient.Client.ReceiveBufferSize = DEVICE_SOCKBUF_DATA;

			// connect to the device
			try {
				udpClient.Connect(Defines.DEVICE_IPADDR, Defines.DEVICE_PORT_CTRL_DEVICE);
				tcpClient.Connect(Defines.DEVICE_IPADDR, Defines.DEVICE_PORT_DATARECV);
			}
			catch (Exception e)
			{
				// connect error
				Console.WriteLine("device connect error ...");
				Console.WriteLine(e.Message);

				udpClient.Close();
				tcpClient.Close();

				comRef.cmdClient = null;
				comRef.dataClient = null;

				return(-1);
			}

			// memorize instances for future reference
			comRef.cmdClient = udpClient;
			comRef.dataClient = tcpClient;

			return(0);
		}


		//#########################################################################
		//  Name	  : sitcp_disconnect
		//  Function  : disconnect device
		//  Arguments :
		//	  in comRef : SiTCPcom class instance
		//  Return : none
		//  Note :
		//#########################################################################
		public static void sitcp_disconnect(SiTCPcom comRef)
		{
			// close UDP connection and dispose UdpClient instance
			if (comRef.cmdClient.Client.Connected == true) {
				(comRef.cmdClient).Close();
			}

			// close TCP connection and dispose TcpClient instance
			if (comRef.dataClient.Connected == true) {
				(comRef.dataClient).Close();
			}
		}


		//#########################################################################
		//  Name	  : sitcp_read
		//  Function  : command communication(read request)
		//  Arguments :
		//	  in  cmdClient : instance of UdpClient class for command communication
		//	  in  cmd_addr  : command address
		//	  in  size	    : size of data to be read(byte)
		//	  out ackdt	    : response data
		//  Return :
		//	   0 : success
		//	  -1 : no response within a period of timeout value
		//	  -2 : invalid size of received packet
		//	  -3 : received packet is not a response to the request
		//  Note : This function doesn't convert received data. It's up to caller.
		//#########################################################################
		static byte r_gid = 0;	// global ID to distinguish read response
		private static int sitcp_read(UdpClient cmdClient, uint cmd_addr, int size, out SITCP_FORMAT ackdt)
		{
			SITCP_FORMAT reqdt, rcvdt;
			SITCP_HEADER cmpdt;
			int result = 0;
			int rbyte;
			int ssize, rsize;
			int sizeSH = Marshal.SizeOf(typeof(SITCP_HEADER));
			int sizeSF = Marshal.SizeOf(typeof(SITCP_FORMAT));
			byte wid;
			byte[] send_buff = new byte[sizeSF];
			IntPtr pt = IntPtr.Zero;

			wid = r_gid++;

			// clear area for reception
			rcvdt = new SITCP_FORMAT();

			// set size of packet for request and response
			ssize = sizeSH;
			rsize = sizeSH + size;

			// construct packet for read request
			reqdt = new SITCP_FORMAT();
			reqdt.header.ver_type = SITCP_VER_TYPE;
			reqdt.header.cmd_flag = SITCP_READ_REQ_CMDFLAG;
			reqdt.header.id = wid;
			reqdt.header.length = (byte)(size);
			reqdt.header.address = (uint)IPAddress.HostToNetworkOrder( (int)cmd_addr );

			// convert the packet to array of bytes
			try {
				pt = Marshal.AllocHGlobal(sizeSF);
				Marshal.StructureToPtr(reqdt, pt, false);
				Marshal.Copy(pt, send_buff, 0, sizeSF);
			}
			finally {
				if (pt != IntPtr.Zero) {
					Marshal.FreeHGlobal(pt);
				}
			}

			// send the packet to the target
			cmdClient.Send(send_buff, ssize);

			// receive response
			rbyte = sitcp_recv_cmd(cmdClient, ref rcvdt, DEVICE_TIMEOUT_CMD);

			if (rbyte >= rsize) {
				// construct header part of response packet from request one
				// by change flag, which is only difference, from 'request' to 'acknowledge'
				cmpdt = reqdt.header;
				cmpdt.cmd_flag = SITCP_READ_ACK_CMDFLAG;

				// check if header part is expected one
				if (sitcp_hdr_cmp(cmpdt, rcvdt.header, sizeSH) != true) {
					result = -3;	// not a response to the request
				}

				// convert command(address) from network byte order to host byte order
				rcvdt.header.address = (uint)IPAddress.NetworkToHostOrder( (int)(rcvdt.header.address) );
			}
			else {	// short of data size
				result = -2;
			}

			// pass response data to caller
			ackdt = rcvdt;

			return(result);
		}

		// memcmp( (void *)tgt1, (void *)tgt2, sizeof(tgt1) )
		private static bool sitcp_hdr_cmp(SITCP_HEADER tgt1, SITCP_HEADER tgt2, int hsize)
		{
			// allocate memory from the unmanaged memory of the process

			// convert tgt1 to array of bytes
			IntPtr pt1 = Marshal.AllocHGlobal(hsize);
			byte[] buff1 = new byte[hsize];
			Marshal.StructureToPtr(tgt1, pt1, false);
			Marshal.Copy(pt1, buff1, 0, hsize);

			// convert tgt2 to array of bytes
			IntPtr pt2 = Marshal.AllocHGlobal(hsize);
			byte[] buff2 = new byte[hsize];
			Marshal.StructureToPtr(tgt2, pt2, false);
			Marshal.Copy(pt2, buff2, 0, hsize);

			// compare tgt1 and tgt2 as array of bytes
			bool ret = buff1.SequenceEqual(buff2);

			Marshal.FreeHGlobal(pt1);
			Marshal.FreeHGlobal(pt2);

			return(ret);
		}


		//#########################################################################
		//  Name	  : sitcp_read_16bit
		//  Function  : command communication(read request for 16bit value)
		//  Arguments :
		//	  in  comRef   : SiTCPcom class instance for referring UdpClient
		//	  in  cmd_addr : command address
		//	  out errcd	   : error info; see return value of sitcp_read
		//  Return : read value or 0(in case of error)
		//  Note : This function converts received data from network byte order
		//		 to host byte order
		//#########################################################################
		public static ushort sitcp_read_16bit(SiTCPcom comRef, uint cmd_addr, out int errcd)
		{
			SITCP_FORMAT rdat;
			int ret;
			ushort result = 0;

			ret = sitcp_read(comRef.cmdClient, cmd_addr, sizeof(ushort), out rdat);
			if (ret == 0) {
				result = (ushort)IPAddress.NetworkToHostOrder( (short)rdat.u16[0] );
			}

			errcd = ret;

			return(result);
		}


		//#########################################################################
		//  Name	  : sitcp_read_32bit
		//  Function  : command communication(read request for 32bit value)
		//  Arguments :
		//	  in  comRef   : SiTCPcom class instance for referring UdpClient
		//	  in  cmd_addr : command address
		//	  out errcd	   : error info; see return value of sitcp_read
		//  Return : read value or 0(in case of error)
		//  Note : This function converts received data from network byte order
		//		 to host byte order
		//#########################################################################
		public static uint sitcp_read_32bit(SiTCPcom comRef, uint cmd_addr, out int errcd)
		{
			SITCP_FORMAT rdat;
			int ret;
			uint result = 0;

			ret = sitcp_read(comRef.cmdClient, cmd_addr, sizeof(uint), out rdat);
			if (ret == 0) {
				result = (uint)IPAddress.NetworkToHostOrder( (int)rdat.u32[0] );
			}

			errcd = ret;

			return(result);
		}


		//#########################################################################
		//  Name	  : sitcp_read_48bit
		//  Function  : command communication(read request for 48bit value)
		//  Arguments :
		//	  in  comRef   : SiTCPcom class instance for referring UdpClient
		//	  in  cmd_addr : command address
		//	  out errcd	   : error info; see return value of sitcp_read
		//  Return : read value or 0(in case of error)
		//  Note : This function converts received data from network byte order
		//		 to host byte order
		//#########################################################################
		public static ulong sitcp_read_48bit(SiTCPcom comRef, uint cmd_addr, out int errcd)
		{
			SITCP_FORMAT rdat;
			int ret, i;
			byte[] abyte = new byte[8];
			ulong result = 0;
			ulong conved;

			ret = sitcp_read(comRef.cmdClient, cmd_addr, 6, out rdat);
			if (ret == 0) {
				// copy upper 6 bytes to array abyte(start from 2nd index)
				for (i = 0; i < 6; i++) {
					abyte[i+2] = rdat.u08[i];
				}

				// convert byte array to ulong
				conved = BitConverter.ToUInt64(abyte, 0);
				result = (ulong)IPAddress.NetworkToHostOrder( (long)conved );
			}

			errcd = ret;

			return(result);
		}


		//#########################################################################
		//  Name	  : sitcp_read_64bit
		//  Function  : command communication(read request for 64bit value)
		//  Arguments :
		//	  in  comRef   : SiTCPcom class instance for referring UdpClient
		//	  in  cmd_addr : command address
		//	  out errcd	   : error info; see return value of sitcp_read
		//  Return : read value or 0(in case of error)
		//  Note : This function converts received data from network byte order
		//		 to host byte order
		//#########################################################################
		public static ulong sitcp_read_64bit(SiTCPcom comRef, uint cmd_addr, out int errcd)
		{
			SITCP_FORMAT rdat;
			int ret;
			ulong result;

			result = (ulong)(0);

			ret = sitcp_read(comRef.cmdClient, cmd_addr, sizeof(ulong), out rdat);
			if (ret == 0) {
				result = (ulong)IPAddress.NetworkToHostOrder( (long)rdat.u64[0] );
			}

			errcd = ret;

			return(result);
		}


		//#########################################################################
		//  Name	  : sitcp_write
		//  Function  : command communication(write request)
		//  Arguments :
		//	  in  cmdClient : instance of UdpClient class for command communication
		//	  in  cmd_addr  : command address
		//	  in  wbuffer   : data to be written 
		//	  in  size	    : size of data to be written(byte)
		//  Return :
		//	   0 : success
		//	  -2 : invalid size of received packet
		//	  -3 : received packet is not a response to the request
		//  Note : This function doesn't convert written data. It's up to caller.
		//#########################################################################
		static byte w_gid = 0;
		private static int sitcp_write(UdpClient cmdClient, uint cmd_addr, byte[] wbuffer, int size)
		{
			SITCP_FORMAT reqdt, rcvdt;
			SITCP_HEADER cmpdt;
			int result = 0;
			int rbyte;
			int ssize, rsize;
			int sizeSH = Marshal.SizeOf(typeof(SITCP_HEADER));
			int sizeSF = Marshal.SizeOf(typeof(SITCP_FORMAT));
			byte wid;
			byte[] send_buff = new byte[sizeSF];
			IntPtr pt = IntPtr.Zero;

			wid = w_gid++;

			// clear area for reception
			rcvdt = new SITCP_FORMAT();

			// set size of packet for request and response
			ssize = sizeSH + size;
			rsize = ssize;

			// construct packet for write request
			reqdt = new SITCP_FORMAT();
			reqdt.header.ver_type = SITCP_VER_TYPE;
			reqdt.header.cmd_flag = SITCP_WRITE_REQ_CMDFLAG;
			reqdt.header.id = wid;
			reqdt.header.length = (byte)(size);
			reqdt.header.address = (uint)IPAddress.HostToNetworkOrder( (int)cmd_addr );

			// convert the packet to array of bytes
			try {
				// header part
				pt = Marshal.AllocHGlobal(sizeSF);
				Marshal.StructureToPtr(reqdt, pt, false);
				Marshal.Copy(pt, send_buff, 0, sizeSH);

				Marshal.FreeHGlobal(pt);

				// data part
				pt = Marshal.AllocHGlobal(size);
				Marshal.Copy(wbuffer, 0, pt, size);			// copy wbuffer from index 0 to size
				Marshal.Copy(pt, send_buff, sizeSH, size);	// copy to index sizeSH of send_buff
			}
			finally {
				if (pt != IntPtr.Zero) {
					Marshal.FreeHGlobal(pt);
				}
			}

			// send the packet to the target
			cmdClient.Send(send_buff, ssize);

			// receive response
			rbyte = sitcp_recv_cmd(cmdClient, ref rcvdt, DEVICE_TIMEOUT_CMD);

			if (rbyte >= rsize) {
				// construct header part of response packet from request one
				// by change flag, which is only difference, from 'request' to 'acknowledge'
				cmpdt = reqdt.header;
				cmpdt.cmd_flag = SITCP_WRITE_ACK_CMDFLAG;

				// check if header part is expected one
				if (sitcp_hdr_cmp(cmpdt, rcvdt.header, sizeSH) != true) {
					result = -3;	// not a response to the request
				}

				// convert command(address) from network byte order to host byte order
				rcvdt.header.address = (uint)IPAddress.NetworkToHostOrder( (int)rcvdt.header.address );
			}
			else { // short of data size
				result = -2;
			}

			return(result);
		}


		//#########################################################################
		//  Name	  : sitcp_write_16bit
		//  Function  : command communication(write request for 16bit value)
		//  Arguments :
		//	  in comRef   : SiTCPcom class instance for referring UdpClient
		//	  in cmd_addr : command address
		//	  in wdata	  : data to be written
		//  Return : same as sitcp_write
		//  Note : This function converts target data from host byte order
		//		 to network byte order
		//#########################################################################
		public static int sitcp_write_16bit(SiTCPcom comRef, uint cmd_addr, ushort wdata)
		{
			int result;
			ushort value;
			Command cmd_ref = new Command();

			value = (ushort)IPAddress.HostToNetworkOrder( (short)wdata );
			byte[] byt_arr = BitConverter.GetBytes(value);
			result = sitcp_write(comRef.cmdClient, cmd_addr, byt_arr, byt_arr.Length);

			// In case of CH specific command, it takes time to reflect the value to the device,
			// so wait here for a while in order to avoid premature consecutive command.
			if (cmd_addr >= cmd_ref.GetVal("COMMAND_ACG")) {
				Thread.Sleep((int)Defines.CH_CMD_INTERVAL);
			}

			return(result);
		}


		//#########################################################################
		//  Name	  : sitcp_write_32bit
		//  Function  : command communication(write request for 32bit value)
		//  Arguments :
		//	  in comRef   : SiTCPcom class instance for referring UdpClient
		//	  in cmd_addr : command address
		//	  in wdata	  : data to be written
		//  Return : same as sitcp_write
		//  Note : This function converts target data from host byte order
		//		 to network byte order
		//#########################################################################
		public static int sitcp_write_32bit(SiTCPcom comRef, uint cmd_addr, uint wdata)
		{
			int result;
			uint value;

			value = (uint)IPAddress.HostToNetworkOrder( (int)wdata );
			byte[] byt_arr = BitConverter.GetBytes(value);
			result = sitcp_write(comRef.cmdClient, cmd_addr, byt_arr, byt_arr.Length);

			return(result);
		}


		//#########################################################################
		//  Name	  : sitcp_write_48bit
		//  Function  : command communication(write request for 48bit value)
		//  Arguments :
		//	  in comRef   : SiTCPcom class instance for referring UdpClient
		//	  in cmd_addr : command address
		//	  in wdata	  : data to be written
		//  Return : same as sitcp_write
		//  Note : This function converts target data from host byte order
		//		 to network byte order
		//#########################################################################
		public static int sitcp_write_48bit(SiTCPcom comRef, uint cmd_addr, ulong wdata)
		{
			int result, i;
			byte[] byt_arr = new byte[8];
			ulong wvalue;

			wvalue = (ulong)IPAddress.HostToNetworkOrder( (long)wdata );
			byte[] abyte = BitConverter.GetBytes(wvalue);

			// copy lower 6 bytes to byt_arr(from upper byte)
			for (i = 0; i < 6; i++) {
				byt_arr[i] = abyte[i+2] ;
			}

			result = sitcp_write(comRef.cmdClient, cmd_addr, byt_arr, byt_arr.Length);

			return(result);
		}


		//#########################################################################
		//  Name	  : sitcp_write_64bit
		//  Function  : command communication(write request for 64bit value)
		//  Arguments :
		//	  in comRef   : SiTCPcom class instance for referring UdpClient
		//	  in cmd_addr : command address
		//	  in wdata	  : data to be written
		//  Return : same as sitcp_write
		//  Note : This function converts target data from host byte order
		//		 to network byte order
		//#########################################################################
		public static int sitcp_write_64bit(SiTCPcom comRef, uint cmd_addr, ulong wdata)
		{
			int result;
			ulong value;

			value = (ulong)IPAddress.HostToNetworkOrder( (long)wdata );
			byte[] byt_arr = BitConverter.GetBytes(value);
			result = sitcp_write(comRef.cmdClient, cmd_addr, byt_arr, byt_arr.Length);

			return(result);
		}


		//#########################################################################
		//  Name	  : sitcp_recv_cmd
		//  Function  : command communication(receive response)
		//  Arguments :
		//	  in  cmdClient : instance of UdpClient class for command communication
		//	  out rbuffer   : buffer for received data
		//	  in  timeout   : receive timeout value(msec)
		//  Return : 
		//	  positive value: received data size
		//	  -11 : no response within a period of timeout value
		//	  -12 : no data
		//  Note :
		//#########################################################################
		private static int sitcp_recv_cmd(UdpClient cmdClient, ref SITCP_FORMAT rbuffer, int timeout)
		{
			int result = 0;

			// create an IPEndPoint instance that represents a remote host from which the data was sent
			IPEndPoint remoteEP = new IPEndPoint( IPAddress.Parse(Defines.DEVICE_IPADDR),
													 Defines.DEVICE_PORT_CTRL_DEVICE );

			// set timeout value
			cmdClient.Client.ReceiveTimeout = timeout;

			try {
				// wait until received data is ready
				byte[] receiveBytes = cmdClient.Receive(ref remoteEP);

				result = receiveBytes.Length;
				if (result != 0) {
					// copy the array data(start from index 0) to unmanaged memory pointer
					IntPtr toRbuffer = Marshal.AllocHGlobal(result);
					Marshal.Copy(receiveBytes, 0, toRbuffer, result);

					// convert it to SITCP_FORMAT then store into rbuffer;
					rbuffer = (SITCP_FORMAT)
								Marshal.PtrToStructure( toRbuffer, typeof(SITCP_FORMAT) );
				} else {
					result = -12;
				}
			}
			catch(SocketException) {
				// presume timeout; ErrorCode = 10060(WSAETIMEDOUT)
				result = -11;
			}

			return(result);
		}


		//#########################################################################
		//  Name	  : sitcp_recv_data
		//  Function  : receive data by fast data transfer(TCP)
		//  Arguments :
		//	  in comRef  : instance of SiTCPcom class
		//	  in rbuffer : buffer for received data
		//	  in idx     : index of rbuffer where data storing starts
		//	  in size	 : size of data to be received(byte)
		//	  in timeout : receive timeout value(msec)
		//  Return : 
		//	  positive value: received data size
		//	  -1 : communication error
		//  Note :
		//#########################################################################
		const int TMP_BUF_SIZE = 2048;
		public static int sitcp_recv_data(SiTCPcom comRef, ref byte[] rbuffer, int idx,
															int size, int timeout)
		{
			int tval, tval2nd;
			int result;
			int rbyte, rsize, req_size;
			ulong count;

			result = 0;
			count = 0;

			// save timeout value for 1st read
			tval = timeout;
			// save timeout value for 2nd read or later
			tval2nd = timeout * 10;

			// create NetworkStream for receiving data
			NetworkStream ns = comRef.dataClient.GetStream();
			// create temporary buffer to store receive data 
			byte[] tmp_buff = new byte[TMP_BUF_SIZE];

			do {
				if ((count <= 1) && (timeout > 0)) {
					// set timeout value
					ns.ReadTimeout = tval;

					// update timeout value for 2nd or later read
					tval = tval2nd;
				}

				// update request data size which fits temporary buffer
				rsize = size - result;
				req_size = (rsize < TMP_BUF_SIZE ) ? rsize : TMP_BUF_SIZE;

				try {
					// receive data
					rbyte = ns.Read(tmp_buff, 0, req_size);
				}
				catch (IOException) {
					// presume timeout
					break;
				}

				if (rbyte > 0) {
					// copy received data to specified buffer
					// (start from index 'sum of idx and result' of rbuffer)
					Array.Copy(tmp_buff, 0, rbuffer, idx + result, rbyte);

					// update total received data size
					result += rbyte;

					if (result >= size) {
						// finish when total received data size reached to request one
						break;
					}
				}
				else { // no data
					result = -1;
					break;
				}

				count++;
			} while(ns.DataAvailable);

			return(result);
		}


	}	// end of class
}	// end of namespace

